# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0+
"""
The "resultsdb handler".

This module is responsible for listening new results from ResultsDB. When a new
result is received, Greenwave will check all applicable policies for that item,
and if the new result causes the decision to change it will publish a message
to the message bus about the newly satisfied/unsatisfied policy.
"""

import collections
import logging
import re

from flask import current_app
import fedmsg.consumers
import requests

import greenwave.app_factory
import greenwave.resources
from greenwave.api_v1 import subject_type_identifier_to_list
from greenwave.monitoring import publish_decision_exceptions_result_counter
from greenwave.policies import RemoteRule
from greenwave.safe_yaml import SafeYAMLError

import xmlrpc.client


log = logging.getLogger(__name__)


def _guess_product_version(toparse, koji_build=False):
    if toparse == 'rawhide' or toparse.startswith('Fedora-Rawhide'):
        return 'fedora-rawhide'

    product_version = None
    if toparse.startswith('f') and koji_build:
        product_version = 'fedora-'
    elif toparse.startswith('epel'):
        product_version = 'epel-'
    elif toparse.startswith('el'):
        product_version = 'rhel-'
    elif toparse.startswith('fc') or toparse.startswith('Fedora'):
        product_version = 'fedora-'

    if product_version:
        # seperate the prefix from the number
        result = list(filter(None, '-'.join(re.split(r'(\d+)', toparse)).split('-')))
        if len(result) >= 2:
            try:
                int(result[1])
                product_version += result[1]
                return product_version
            except ValueError:
                pass

    return None


def _subject_product_version(subject_identifier, subject_type):
    if subject_type == 'koji_build':
        try:
            short_prod_version = subject_identifier.split('.')[-1]
            return _guess_product_version(short_prod_version, koji_build=True)
        except KeyError:
            pass

    if subject_type == "compose":
        return _guess_product_version(subject_identifier)

    koji_base_url = current_app.config['KOJI_BASE_URL']
    if koji_base_url:
        proxy = xmlrpc.client.ServerProxy(koji_base_url)
        try:
            build = proxy.getBuild(subject_identifier)
            if build:
                target = proxy.getTaskRequest(build['task_id'])[1]
                return _guess_product_version(target, koji_build=True)
        except KeyError:
            pass
        except xmlrpc.client.Fault:
            pass


def _invalidate_results_cache(
        cache, subject_type, subject_identifier, testcase):
    """
    Removes results for given parameters from cache.
    """
    key = greenwave.resources.results_cache_key(
        subject_type, subject_identifier, testcase)

    log.debug("Invalidating cache for %r", key)

    try:
        cache.delete(key)
    except KeyError:
        log.debug("No cache value found for %r", key)

    # Also invalidate query results without test case name.
    if testcase:
        _invalidate_results_cache(
            cache, subject_type, subject_identifier, testcase=None)


class ResultsDBHandler(fedmsg.consumers.FedmsgConsumer):
    """
    Handle a new result.

    Attributes:
        topic (list): A list of strings that indicate which fedmsg topics this consumer listens to.
    """

    config_key = 'resultsdb_handler'

    def __init__(self, hub, *args, **kwargs):
        """
        Initialize the ResultsDBHandler, subscribing it to the appropriate topics.

        Args:
            hub (moksha.hub.hub.CentralMokshaHub): The hub from which this handler is consuming
                messages. It is used to look up the hub config.
        """

        prefix = hub.config.get('topic_prefix')
        env = hub.config.get('environment')
        suffix = hub.config.get('resultsdb_topic_suffix', 'taskotron.result.new')
        self.topic = ['.'.join([prefix, env, suffix])]
        self.fedmsg_config = fedmsg.config.load_config()

        super(ResultsDBHandler, self).__init__(hub, *args, **kwargs)

        self.flask_app = greenwave.app_factory.create_app()
        self.cache = self.flask_app.cache

        log.info('Greenwave resultsdb handler listening on: %s', self.topic)

    @staticmethod
    def announcement_subjects(message):
        """
        Yields pairs of (subject type, subject identifier) for announcement
        consideration from the message.

        Args:
            message (munch.Munch): A fedmsg about a new result.
        """

        try:
            data = message['msg']['data']  # New format
        except KeyError:
            data = message['msg']['task']  # Old format

        def _decode(value):
            """ Decode either a string or a list of strings. """
            if value and len(value) == 1:
                value = value[0]
            return value

        _type = _decode(data.get('type'))
        if _type == 'bodhi_update' and 'item' in data:
            yield ('bodhi_update', _decode(data['item']))
        if _type == 'compose' and 'item' in data:
            yield ('compose', _decode(data['item']))
        if 'productmd.compose.id' in data:
            yield ('compose', _decode(data['productmd.compose.id']))
        if (_type == 'koji_build' and 'item' in data or
                _type == 'brew-build' and 'item' in data or
                'original_spec_nvr' in data):
            if _type in ['koji_build', 'brew-build']:
                nvr = _decode(data['item'])
            else:
                nvr = _decode(data['original_spec_nvr'])
            yield ('koji_build', nvr)

    def consume(self, message):
        """
        Process the given message and take action.

        Args:
            message (munch.Munch): A fedmsg about a new result.
        """
        message = message.get('body', message)
        log.debug('Processing message "%s"', message)

        try:
            testcase = message['msg']['testcase']['name']  # New format
        except KeyError:
            testcase = message['msg']['task']['name']  # Old format

        try:
            result_id = message['msg']['id']  # New format
        except KeyError:
            result_id = message['msg']['result']['id']  # Old format

        with self.flask_app.app_context():
            for subject_type, subject_identifier in self.announcement_subjects(message):
                log.debug('Considering subject %s: %r', subject_type, subject_identifier)
                _invalidate_results_cache(
                    self.cache, subject_type, subject_identifier, testcase)
                self._publish_decision_changes(subject_type, subject_identifier,
                                               result_id, testcase)

    @publish_decision_exceptions_result_counter.count_exceptions()
    def _publish_decision_changes(self, subject_type, subject_identifier, result_id, testcase):
        """
        Process the given subject and publish a message if the decision is changed.

        Args:
            subject (munch.Munch): A subject argument, used to query greenwave.
            result_id (int): A result ID to ignore for comparison.
            testcase (munch.Munch): The name of a testcase to consider.
        """
        # Also need to apply policies for each build in the update.
        if subject_type == 'bodhi_update':
            subject_types = set([subject_type, 'koji_build'])
        else:
            subject_types = set([subject_type])

        # Build a set of all policies which might apply to this new results
        applicable_policies = set()
        remote_policies = set()
        for policy in current_app.config['policies']:
            for rule in policy.rules:
                if isinstance(rule, RemoteRule):
                    try:
                        for remote_policy in rule.get_policies(policy, subject_identifier):
                            remote_policy.product_versions = set(
                                remote_policy.product_versions).intersection(set(
                                    policy.product_versions))
                            if (remote_policy.product_versions and
                                    remote_policy.decision_context == policy.decision_context):
                                remote_policies = remote_policies.union(set([remote_policy]))
                    except SafeYAMLError as e:
                        pass

        tmp_policies = remote_policies.union(current_app.config['policies'])
        for policy in tmp_policies:
            if policy.subject_type in subject_types:
                testcases = (
                    getattr(rule, 'test_case_name', None)
                    for rule in policy.rules)

                if testcase in testcases:
                    applicable_policies.add(policy)

        log.debug("messaging: found %i applicable policies of %i for testcase %r",
                  len(applicable_policies), len(current_app.config['policies']), testcase)

        # Given all of our applicable policies, build a map of all decision
        # context we know about, and which product versions they relate to.
        decision_contexts = collections.defaultdict(set)
        product_version = _subject_product_version(subject_identifier, subject_type)
        for policy in applicable_policies:
            if not product_version or policy.applies_to_product_version(product_version):
                if product_version:
                    versions = set([product_version])
                else:
                    versions = set(policy.product_versions)
                decision_contexts[policy.decision_context].update(versions)
        log.debug("messaging: found %i decision contexts", len(decision_contexts))

        # For every context X version combination, ask greenwave if this new
        # result pushes any decisions over a threshold.
        for decision_context in sorted(decision_contexts.keys()):
            product_versions = decision_contexts[decision_context]
            for product_version in sorted(product_versions):
                greenwave_url = self.fedmsg_config['greenwave_api_url'] + '/decision'

                data = {
                    'decision_context': decision_context,
                    'product_version': product_version,
                    'subject_type': subject_type,
                    'subject_identifier': subject_identifier,
                }

                try:
                    decision = greenwave.resources.retrieve_decision(greenwave_url, data)

                    # get old decision
                    data.update({
                        'ignore_result': [result_id],
                    })
                    old_decision = greenwave.resources.retrieve_decision(greenwave_url, data)
                except requests.exceptions.HTTPError as e:
                    log.exception('Failed to retrieve decision for data=%s, error: %s', data, e)
                    continue

                if decision == old_decision:
                    log.debug('Skipped emitting fedmsg, decision did not change: %s', decision)
                else:
                    decision.update({
                        'subject_type': subject_type,
                        'subject_identifier': subject_identifier,
                        # subject is for backwards compatibility only:
                        'subject': subject_type_identifier_to_list(subject_type,
                                                                   subject_identifier),
                        'decision_context': decision_context,
                        'product_version': product_version,
                        'previous': old_decision,
                    })
                    log.debug('Emitted a fedmsg, %r, on the "%s" topic', decision,
                              'greenwave.decision.update')
                    fedmsg.publish(topic='decision.update', msg=decision)
